import os

ENVS_DIRNAME = 'envs'
ENV_FILE_EXTENSION = '.json'
SKALE_ENVS_FILENAME = 'envs.yml'


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
PROJECT_DIR = os.path.join(CURRENT_DIR, os.pardir)

ENVS_DIR = os.path.join(CURRENT_DIR, ENVS_DIRNAME)
SKALE_ENVS_FILEPATH = os.path.join(ENVS_DIR, SKALE_ENVS_FILENAME)



from skale.contracts.base_contract import BaseContract

from skale.contracts.manager import Manager
from skale.contracts.token import Token
from skale.contracts.schains import SChains
from skale.contracts.nodes import Nodes
from skale.contracts.groups import Groups
from skale.contracts.validators import Validators



from skale.contracts import BaseContract
from skale.utils.helper import sign_and_send


class Validators(BaseContract):

    def get_periods(self, account):
        return self.contract.functions.getPeriods().call({'from': account})

    def get_validated_array(self, node_id, account):
        return self.contract.functions.getValidatedArray(node_id).call({'from': account})

    def add_authorized(self, address, wallet):
        op = self.contract.functions.addAuthorized(address)
        tx = sign_and_send(self.skale, op, 100000, wallet)
        return {'tx': tx}


from skale.contracts import BaseContract
from skale.utils.helper import format, ip_from_bytes
from skale.utils.helper import sign_and_send

FIELDS = [
    'name',
    'owner',
    'indexInOwnerList',
    #'storageBytes',
    #'cpu',
    #'transactionThroughput',
    'lifetime',
    'startDate',
    'numberOfNodes',
    'deposit',
    'groupIndex'
    #'nodes'
]


class SChains(BaseContract):
    def __get_raw(self, name):
        return self.contract.functions.schains(name).call()

    @format(FIELDS)
    def get(self, id):
        return self.__get_raw(id)

    @format(FIELDS)
    def get_by_name(self, name):
        id = self.get_id_by_name(name)
        return self.__get_raw(id)

    def __get_schain_nodes_raw(self, schain_name):
        return self.contract.functions.getSchainNodes(schain_name).call()

    def get_schain_nodes(self, schain_name):
        nodes = []
        raw_nodes = self.__get_schain_nodes_raw(schain_name)

        for i, node in enumerate(raw_nodes):
            node_id_bytes = node[0:10]
            node_id = int.from_bytes(node_id_bytes, byteorder='big')

            ip_bytes = node[10:14]
            ip = ip_from_bytes(ip_bytes)

            port_bytes = node[14:16]
            port = int.from_bytes(port_bytes, byteorder='big')

            nodes.append({
                'schainIndex': i,
                'nodeID': node_id,
                'ip': ip,
                'basePort': port
            })
        return nodes

    def get_schain_list_size(self, account):
        return self.contract.functions.getSchainListSize().call({'from': account})

    def get_schain_by_index(self, index):
        return self.contract.functions.getSchainByIndex(index).call()

    # todo: doesn't work now
    def get_schain_name_by_schain_id(self, schain_id):
        raise NotImplementedError
        # return self.contract.functions.getSchainNameBySchainId(schain_id).call()

    def get_id_by_name(self, name):
        return self.contract.functions.getSchainIdBySchainName(name).call()

    def get_schain_ids_for_node(self, node_id):
        # return self.contract.functions.schainsForNodes(int(node_id)).call() # todo: new function - problem with bytes32 map
        return self.contract.functions.getSchainIdsForNode(node_id).call()  # todo: function will be removed

    def get_schains_for_node(self, node_id):
        schains = []
        schain_contract = self.skale.get_contract_by_name('schains')
        schain_ids = self.get_schain_ids_for_node(node_id)
        for schain_id in schain_ids:
            # name = self.get_schain_name_by_schain_id(schain_id)
            schain = schain_contract.get(schain_id)
            schains.append(schain)
        return schains

    def add_authorized(self, address, wallet):
        op = self.contract.functions.addAuthorized(address)
        tx = sign_and_send(self.skale, op, 100000, wallet)
        return {'tx': tx}

    def get_schain_price(self, indexOfType):
        return self.contract.functions.getSchainPrice(indexOfType).call()


import json, socket, os
from random import randint
import ipaddress
from time import sleep
import logging
import yaml
import string, random

from skale.config import SKALE_ENVS_FILEPATH, ENVS_DIR, ENV_FILE_EXTENSION
from skale.blockchain_env import BlockchainEnv

logger = logging.getLogger(__name__)


def format(fields):
    """
        Transform array to object with passed fields
        Usage:
        @format(['field_name1', 'field_name2'])
        def my_method()
            return [0, 'Test']

        => {'field_name1': 0, 'field_name2': 'Test'}
    """

    def real_decorator(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            obj = {}
            for i, field in enumerate(fields):
                obj[field] = result[i]
            return obj

        return wrapper

    return real_decorator


def get_receipt(web3, tx):
    return web3.eth.getTransactionReceipt(tx)


def get_eth_nonce(web3, address):
    return web3.eth.getTransactionCount(address)


def get_nonce(skale, address):
    lib_nonce = skale.nonces.get(address)
    if not lib_nonce:
        lib_nonce = get_eth_nonce(skale.web3, address)
        skale.nonces.get(address)
    else:
        lib_nonce += lib_nonce
    return lib_nonce


def sign_and_send(skale, method, gas_amount, wallet):
    eth_nonce = get_nonce(skale, wallet['address'])
    txn = method.buildTransaction({
        'gas': gas_amount,
        'nonce': eth_nonce
    })
    signed_txn = skale.web3.eth.account.signTransaction(txn, private_key=wallet['private_key'])
    tx = skale.web3.eth.sendRawTransaction(signed_txn.rawTransaction)
    logger.info(f'{method.__class__.__name__} - transaction_hash: {skale.web3.toHex(tx)}')
    return tx


def await_receipt(web3, tx, retries=10, timeout=5):
    for _ in range(0, retries):
        receipt = get_receipt(web3, tx)
        if (receipt != None):
            return receipt
        sleep(timeout)
    return None


def ip_from_bytes(bytes):
    return socket.inet_ntoa(bytes)


def ip_to_bytes(ip):
    return socket.inet_aton(ip)


def check_port(port):
    if port not in range(1, 65535):
        raise ValueError(f'{port} does not appear to be a valid port. Allowed range: 1-65535')


def check_ip(ip):
    return ipaddress.ip_address(ip)


def generate_ws_addr(ip, port):
    check_ip(ip)
    check_port(port)
    return 'ws://' + ip + ':' + str(port)

def get_contracts_data_filepath(skale_env):
    return os.path.join(ENVS_DIR, f'{skale_env.value}.json')

def get_abi(skale_env, abi_filepath=None):
    if (skale_env != BlockchainEnv.CUSTOM):
        abi_filepath = get_contracts_data_filepath(skale_env)
    with open(abi_filepath) as data_file:
        return json.load(data_file)

def generate_nonce():
    return randint(0, 65534)


def random_string(size=6, chars=string.ascii_lowercase):
    return ''.join(random.choice(chars) for x in range(size))


def generate_random_ip():
    return '.'.join('%s' % random.randint(0, 255) for i in range(4))


def generate_random_name(len=8):
    return ''.join(random.choices(string.ascii_uppercase + string.digits, k=len))


def generate_random_port():
    return random.randint(0, 60000)

def load_yaml(filepath):
    with open(filepath, 'r') as stream:
        try:
            return yaml.load(stream)
        except yaml.YAMLError as exc:
            print(exc)

def load_skale_env_file():
    return load_yaml(SKALE_ENVS_FILEPATH)

def load_skale_env_config(env):
    env_file = load_skale_env_file()
    return env_file['envs'][env.value]

def generate_custom_config(ip, ws_port):
    if (not ip or not ws_port):
        raise ValueError(
            f'For custom init you should provide ip and ws_port: {ip}, {ws_port}')
    return {
        'ip': ip,
        'ws_port': ws_port,
    }


def send_eth(skale, account, amount, wallet):
    eth_nonce = get_nonce(skale, wallet['address'])
    txn = {
        'to': account,
        'from': wallet['address'],
        'value': amount,
        'gasPrice': skale.web3.eth.gasPrice,
        'gas': 21000,
        'nonce': eth_nonce
    }
    signed_txn = skale.web3.eth.account.signTransaction(txn, private_key=wallet['private_key'])
    tx = skale.web3.eth.sendRawTransaction(signed_txn.rawTransaction)

    logger.info(f'eth sent to {account}, amount: {amount} - transaction_hash: {skale.web3.toHex(tx)}')
    return tx


from web3 import Web3
import socket
import logging

from skale.contracts import BaseContract
import skale.utils.helper as Helper
from skale.utils.helper import sign_and_send

from skale.utils.constants import NODE_DEPOSIT, GAS, OP_TYPES

logger = logging.getLogger(__name__)


class Manager(BaseContract):

    def create_node(self, ip, port, name, wallet):
        logger.info(f'create_node: {ip}:{port}, name: {name}')

        token = self.skale.get_contract_by_name('token')
        skale_nonce = Helper.generate_nonce()
        transaction_data = self.create_node_data_to_bytes(ip, port, name, wallet['address'], skale_nonce)

        op = token.contract.functions.transfer(self.address, NODE_DEPOSIT, transaction_data)
        tx = sign_and_send(self.skale, op, GAS['create_node'], wallet)
        return {'tx': tx, 'nonce': skale_nonce}

    def create_node_data_to_bytes(self, ip, port, name, address, nonce):
        address_fx = Web3.toChecksumAddress(address)[2:]  # fix & cut 0x

        type_bytes = OP_TYPES['create_node'].to_bytes(1, byteorder='big')
        port_bytes = port.to_bytes(2, byteorder='big')
        nonce_bytes = nonce.to_bytes(2, byteorder='big')  # todo
        ip_bytes = socket.inet_aton(ip)
        address_bytes = bytes.fromhex(address_fx)
        name_bytes = name.encode()

        data_bytes = type_bytes + port_bytes + nonce_bytes + ip_bytes + address_bytes + name_bytes
        logger.info(f'create_node_data_to_bytes bytes: {self.skale.web3.toHex(data_bytes)}')

        return data_bytes

    def create_schain(self, lifetime, type_of_nodes, deposit, name, wallet):
        logger.info(f'create_schain: type_of_nodes: {type_of_nodes}, name: {name}')

        token = self.skale.get_contract_by_name('token')
        skale_nonce = Helper.generate_nonce()
        transaction_data = self.create_schain_data_to_bytes(lifetime,
                                                            type_of_nodes, name, skale_nonce)

        op = token.contract.functions.transfer(self.address, deposit, transaction_data)
        tx = sign_and_send(self.skale, op, GAS['create_node'], wallet)
        return {'tx': tx, 'nonce': skale_nonce}

    def create_schain_data_to_bytes(self, lifetime, type_of_nodes, name, nonce):
        type_bytes = OP_TYPES['create_schain'].to_bytes(1, byteorder='big')

        lifetime_hex = lifetime.to_bytes(32, byteorder='big')
        type_of_nodes_hex = type_of_nodes.to_bytes(1, byteorder='big')
        nonce_hex = nonce.to_bytes(2, byteorder='big')
        name_hex = name.encode()

        data_bytes = type_bytes + lifetime_hex + type_of_nodes_hex + nonce_hex + name_hex
        logger.info(f'create_schain_data_to_bytes bytes: {self.skale.web3.toHex(data_bytes)}')
        return data_bytes

    def get_bounty(self, node_id, wallet):
        op = self.contract.functions.getBounty(node_id)
        tx = sign_and_send(self.skale, op, GAS['get_bounty'], wallet)
        return {'tx': tx}

    def send_verdict(self, validator, node_id, downtime, latency, wallet):
        op = self.contract.functions.sendVerdict(validator, node_id, downtime, latency)
        tx = sign_and_send(self.skale, op, GAS['send_verdict'], wallet)
        return {'tx': tx}

    def get_node_next_reward_date(self, node_index):
        return self.contract.functions.getNodeNextRewardDate(node_index).call()



from skale.contracts import BaseContract
from skale.utils.helper import sign_and_send
from skale.utils.constants import GAS

class Token(BaseContract):

    def transfer(self, address, value, wallet):
        op = self.contract.functions.transfer(address, value)
        tx = sign_and_send(self.skale, op, GAS['token_transfer'], wallet)
        return {'tx': tx}

    def get_balance(self, address):
        return self.contract.functions.balanceOf(address).call()

    def add_authorized(self, address, wallet):
        op = self.contract.functions.addAuthorized(address)
        tx = sign_and_send(self.skale, op, GAS['token_transfer'], wallet)
        return {'tx': tx}


from unittest import TestCase
import unittest

from skale import Skale
import skale.utils.helper as Helper


class HelperTest(TestCase):
    def setUp(self):
        self.skale = Skale('51.0.1.99', 8546)

    def test_port_check(self):
        Helper.check_port(3000)
        self.assertRaises(ValueError, Helper.check_port, 70000)

    def test_ip_check(self):
        Helper.check_ip('12.34.56.78')
        self.assertRaises(ValueError, Helper.check_ip, '299.233.354.643')

    def test_generate_nonce(self):
        nonce = Helper.generate_nonce()
        self.assertTrue(1 <= nonce <= 65534)

    def test_get_abi(self):
        default_abi = Helper.get_abi()
        custom_path = Helper.get_default_abipath()
        custom_path_abi = Helper.get_abi(custom_path)
        self.assertEqual(default_abi, custom_path_abi)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
from skale import Skale, EventListener
import skale
from web3 import Web3
import os, random, unittest, socket
from tests._old.utils import init_logger


class ManagerContractTest(TestCase):
    def setUp(self):
        self.skale = Skale('51.0.1.99', 8546)
        init_logger()

    def run_node_listener(self):
        event = skale.nodes.contract.events.NodeCreated
        listener = EventListener(event, self.node_created_handler, 5)
        listener.run()

    def node_created_handler(self, event):
        ip = socket.inet_ntoa(event['args']['ip'])
        config = {
            'node_id': event['args']['nodeIndex'],
            'owner': event['args']['owner'],
            'port': event['args']['port'],
            'time:': event['args']['time'],
            'ip': ip
        }
        # todo: check configs!

    def test_create_node(self):
        ip = '{}.{}.{}.{}'.format(*random.sample(range(0, 255), 4))
        port = random.randint(0, 65000)

        addr = '0x6870ea70c8582a3c3c778ae719b502e4644fd9de'
        address_fx = Web3.toChecksumAddress(addr)

        wallet = {
            'address': address_fx,
            'private_key': os.environ['ETH_PRIVATE_KEY']
        }

        self.skale.manager().create_node(ip, port, wallet)


    def test_get_node(self):
        node_id = 11 # todo: get node_id from the config
        node = self.skale.manager().get_node(node_id)
        print(node)
        # todo: add check



if __name__ == '__main__':
    unittest.main()



import logging

from web3 import Web3, WebsocketProvider

from skale.utils.helper import generate_custom_config, load_skale_env_config, generate_ws_addr, get_abi
import skale.contracts as contracts
from skale.blockchain_env import BlockchainEnv

logger = logging.getLogger(__name__)


class Skale:
    def __init__(self, skale_env, ip=None, ws_port=None, abi_filepath=None):
        env_config = generate_custom_config(ip, ws_port) if (
                    skale_env == BlockchainEnv.CUSTOM) else load_skale_env_config(skale_env)
        logger.info(f'Init skale-py: skale_env: {skale_env}, env_config: {env_config}')
        ws_addr = generate_ws_addr(env_config['ip'], env_config['ws_port'])
        self.web3 = Web3(WebsocketProvider(ws_addr))
        self.abi = get_abi(skale_env, abi_filepath)
        self.__contracts = {}
        self.__init_contracts()
        self.nonces = {}

    def __init_contracts(self):
        self.add_lib_contract('manager', contracts.Manager)
        self.add_lib_contract('token', contracts.Token)
        self.add_lib_contract('schains', contracts.SChains)

        self.add_lib_contract('nodes', contracts.Nodes)
        # self.add_lib_contract('groups', contracts.Groups)
        self.add_lib_contract('validators', contracts.Validators)

    def add_lib_contract(self, name, contract_class):
        address = self.abi[f'skale_{name}_address']
        abi = self.abi[f'skale_{name}_abi']

        self.add_contract(name, contract_class(self, name, address, abi))

    def add_contract(self, name, skale_contract):
        logger.debug(f'Init contract: {name}')
        self.__contracts[name] = skale_contract

    def get_contract_by_name(self, name):
        return self.__contracts[name]

    def __getattr__(self, name):
        if name not in self.__contracts:
            raise AttributeError(name)
        return self.get_contract_by_name(name)



import socket
from Crypto.Hash import keccak
from skale.contracts import BaseContract
from skale.utils.helper import format
from skale.utils.helper import sign_and_send

FIELDS = ['name', 'ip', 'port', 'owner', 'start_date', 'leaving_date', 'last_reward_date', 'second_address']
COMPACT_FIELDS = ['schainIndex', 'nodeID', 'ip', 'basePort']

class Nodes(BaseContract):

    def __get_raw(self, node_id):
        # return self.contract.functions.getNode(node_id).call()
        return self.contract.functions.nodes(node_id).call()

    @format(FIELDS)
    def get(self, node_id):
        return self.__get_raw(node_id)

    def get_active_node_ids(self):
        return self.contract.functions.getActiveNodeIds().call()

    def get_active_node_ips(self):
        return self.contract.functions.getActiveNodeIPs().call()

    def get_active_nodes_by_address(self, account):
        return self.contract.functions.getActiveNodesByAddress().call({'from': account})

    def node_name_to_id(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.hexdigest()

    def is_node_name_available(self, name):
        node_id = self.node_name_to_id(name)
        return not self.contract.functions.nodesNameCheck(node_id).call()

    def is_node_ip_available(self, ip):
        ip_bytes = socket.inet_aton(ip)
        return not self.contract.functions.nodesIPCheck(ip_bytes).call()

    def authorized(self, address):
        return self.contract.functions.authorized(address).call()

    def add_authorized(self, address, wallet):
        op = self.contract.functions.addAuthorized(address)
        tx = sign_and_send(self.skale, op, 100000, wallet)
        return {'tx': tx}


import sys, logging
from logging import Formatter, StreamHandler

def init_logger():
    handlers = []
    formatter = Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    stream_handler = StreamHandler(sys.stderr)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    handlers.append(stream_handler)

    logging.basicConfig(level=logging.DEBUG, handlers=handlers)


NODE_DEPOSIT = 100000000000000000000

GAS = {
    'create_node': 4500000,
    'get_bounty': 2000000,
    'send_verdict': 200000,

    'token_transfer': 30000,
    'add_authorized': 100000
}

OP_TYPES = {
    'create_node': 0x1,
    'create_schain': 0x10
}
